// Copyright (c) 2010 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#define WARNING 1;
#define RAW_LOG(severity, ...);  // Do nothing.
