// Copyright (c) 2012 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include "base/test/mock_devices_changed_observer.h"

namespace base {

MockDevicesChangedObserver::MockDevicesChangedObserver() = default;

MockDevicesChangedObserver::~MockDevicesChangedObserver() = default;

}  // namespace base
